import logging
import numpy as np
from openfisca_tools.data import PrivateDataset, Dataset
from openfisca_uk.data.datasets.frs.frs import FRS
import h5py
from openfisca_uk.data.datasets.frs.enhanced.stages.extension.spi_imputation import (
    impute_incomes,
)
from openfisca_uk.data.datasets.frs.enhanced.utils import (
    clone_and_replace_half,
)
from openfisca_uk.data.datasets.frs.enhanced.stages.extension.uc_transition import (
    migrate_to_universal_credit,
)
from openfisca_uk.data.storage import OPENFISCA_UK_MICRODATA_FOLDER
from time import time


class ExtendedFRS(PrivateDataset):
    name = "extended_frs"
    label = "Extended FRS"
    is_openfisca_compatible = True
    folder_path = OPENFISCA_UK_MICRODATA_FOLDER

    filename_by_year = {
        2022: "extended_frs_2022.h5",
    }

    data_format = Dataset.TIME_PERIOD_ARRAYS

    def generate(self, year: int):
        """Generates the enhanced FRS dataset for OpenFisca-UK.

        Args:
            year (int): The year to generate for (uses the raw FRS from this year).
        """

        logging.info(f"Generating Extended FRS for year {year}")
        frs = FRS.load(year)
        frs_enhanced = h5py.File(self.file(year), mode="w")
        for key in frs.keys():
            frs_enhanced[f"{key}/{year}"] = frs[key][...]
        HOUSEHOLD_COUNT = frs_enhanced[f"household_id/{year}"][...].shape[0]
        frs_enhanced[f"in_original_frs/{year}"] = [True] * HOUSEHOLD_COUNT
        frs_enhanced[f"spi_imputed/{year}"] = [False] * HOUSEHOLD_COUNT
        frs_enhanced[f"uc_migrated/{year}"] = [False] * HOUSEHOLD_COUNT
        frs.close()
        frs_enhanced.close()

        logging.info("Imputing incomes from the SPI")

        pred_income = impute_incomes(year=year)
        clone_and_replace_half(
            self,
            year,
            {
                **{
                    f"{field}/{year}": pred_income[field]
                    for field in pred_income.columns
                },
                f"in_original_frs/{year}": [False] * HOUSEHOLD_COUNT,
                f"spi_imputed/{year}": [True] * HOUSEHOLD_COUNT,
            },
            weighting=0,
        )

        logging.info("Migrating to universal credit")

        uc_migrated = migrate_to_universal_credit(self, year)
        clone_and_replace_half(
            self,
            year,
            {
                **uc_migrated,
                f"in_original_frs/{year}": [False] * HOUSEHOLD_COUNT * 2,
                f"uc_migrated/{year}": [True] * HOUSEHOLD_COUNT * 2,
            },
            weighting=0,
        )

        self.save(
            year,
            f"original_weight/{year}",
            self.load(year, f"household_weight/{year}"),
        )

        logging.info("Finished generating SPI-enhanced FRS.")


ExtendedFRS = ExtendedFRS()



from openfisca_uk.model_api import *
from openfisca_uk.variables.misc.categories.lower_or_higher import (
    LowerOrHigher,
)


class attendance_allowance(Variable):
    value_type = float
    entity = Person
    label = "Attendance Allowance"
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        aa = parameters(period).dwp.attendance_allowance
        category = person("aa_category", period)
        return (
            select(
                [
                    category == LowerOrHigher.HIGHER,
                    category == LowerOrHigher.LOWER,
                    category == LowerOrHigher.NONE,
                ],
                [
                    aa.higher,
                    aa.lower,
                    0,
                ],
            )
            * WEEKS_IN_YEAR
        )


class AA_reported(Variable):
    value_type = float
    entity = Person
    label = "Attendance Allowance (reported)"
    definition_period = YEAR
    unit = GBP


class aa_category(Variable):
    label = "Attendance Allowance category"
    entity = Person
    definition_period = YEAR
    value_type = Enum
    possible_values = LowerOrHigher
    default_value = LowerOrHigher.NONE

    def formula(person, period, parameters):
        aa = parameters(period).dwp.attendance_allowance
        SAFETY_MARGIN = 0.1  # Survey reported values could be slightly below eligible values when they should be above due to data manipulation
        reported_weekly_aa = person("AA_reported", period) / WEEKS_IN_YEAR
        return select(
            [
                reported_weekly_aa >= aa.higher * (1 - SAFETY_MARGIN),
                reported_weekly_aa >= aa.lower * (1 - SAFETY_MARGIN),
                True,
            ],
            [
                LowerOrHigher.HIGHER,
                LowerOrHigher.LOWER,
                LowerOrHigher.NONE,
            ],
        )



from openfisca_uk.model_api import *


class is_blind(Variable):
    label = "Blind"
    documentation = "Certified as blind or severely sight impaired by a consultant ophthalmologist"
    entity = Person
    definition_period = YEAR
    value_type = bool



from openfisca_uk.model_api import *

"""
This file calculates the allowances to which taxpayers are entitled. This follows step 3 of the Income Tax Act 2007 s. 23.
"""


class personal_allowance(Variable):
    value_type = float
    entity = Person
    label = "Personal Allowance for the year"
    unit = GBP
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 35"

    def formula(person, period, parameters):
        PA = parameters(period).tax.income_tax.allowances.personal_allowance
        ANI = person("adjusted_net_income", period)
        excess = max_(0, ANI - PA.maximum_ANI)
        reduction = excess * PA.reduction_rate
        return max_(0, PA.amount - reduction)


class blind_persons_allowance(Variable):
    value_type = float
    entity = Person
    label = "Blind Person's Allowance for the year (not simulated)"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 38"
    unit = GBP


class married_couples_allowance(Variable):
    value_type = float
    entity = Person
    label = "Married Couples' allowance for the year"
    definition_period = YEAR
    unit = GBP


class married_couples_allowance_deduction(Variable):
    value_type = float
    entity = Person
    label = "Deduction from Married Couples' allowance for the year"
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        rate = parameters(
            period
        ).tax.income_tax.allowances.married_couples_allowance.deduction_rate
        return person("married_couples_allowance", period) * rate


class pension_annual_allowance(Variable):
    value_type = float
    entity = Person
    label = "Annual Allowance for pension contributions"
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        allowance = parameters(
            period
        ).tax.income_tax.allowances.annual_allowance
        ANI = person("adjusted_net_income", period)
        reduction = max_(0, ANI - allowance.taper) * allowance.reduction_rate
        return max_(allowance.minimum, allowance.default - reduction)


class trading_allowance(Variable):
    value_type = float
    entity = Person
    label = "Trading Allowance for the year"
    definition_period = YEAR
    reference = "Income Tax (Trading and Other Income) Act 2005 s. 783AF"
    unit = GBP

    def formula(person, period, parameters):
        return parameters(period).tax.income_tax.allowances.trading_allowance


class trading_allowance_deduction(Variable):
    value_type = float
    entity = Person
    label = "Deduction applied by the trading allowance"
    definition_period = YEAR
    reference = "Income Tax (Trading and Other Income) Act 2005 s. 783AF"
    unit = GBP

    def formula(person, period, parameters):
        return min_(
            person("trading_allowance", period),
            person("self_employment_income", period),
        )


class property_allowance(Variable):
    value_type = float
    entity = Person
    label = "Property Allowance for the year"
    definition_period = YEAR
    reference = "Income Tax (Trading and Other Income) Act 2005 s. 783BF"
    unit = GBP

    def formula(person, period, parameters):
        return parameters(period).tax.income_tax.allowances.property_allowance


class property_allowance_deduction(Variable):
    value_type = float
    entity = Person
    label = "Deduction applied by the property allowance"
    definition_period = YEAR
    reference = "Income Tax (Trading and Other Income) Act 2005 s. 783AF"
    unit = GBP

    def formula(person, period, parameters):
        return min_(
            person("property_income", period),
            person("property_allowance", period),
        )


class savings_allowance(Variable):
    value_type = float
    entity = Person
    label = "Savings Allowance for the year"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 12B"
    unit = GBP

    def formula(person, period, parameters):
        tax_band = person("tax_band", period)
        tax_bands = tax_band.possible_values
        amounts = parameters(
            period
        ).tax.income_tax.allowances.personal_savings_allowance
        return select(
            [
                tax_band == tax_bands.ADDITIONAL,
                tax_band == tax_bands.HIGHER,
                (
                    (tax_band == tax_bands.STARTER)
                    | (tax_band == tax_bands.BASIC)
                    | (tax_band == tax_bands.INTERMEDIATE)
                ),
                tax_band == tax_bands.NONE,
            ],
            [amounts.additional, amounts.higher, amounts.basic, amounts.basic],
        )


class dividend_allowance(Variable):
    value_type = float
    entity = Person
    label = "Dividend allowance for the person"
    definition_period = YEAR
    reference = "Income Tax Act 2007 s. 13A"
    unit = GBP

    def formula(person, period, parameters):
        return parameters(period).tax.income_tax.allowances.dividend_allowance


class gift_aid(Variable):
    value_type = float
    entity = Person
    label = "Expenditure under Gift Aid"
    definition_period = YEAR
    unit = GBP


class covenanted_payments(Variable):
    value_type = float
    entity = Person
    label = "Covenanted payments to charities"
    definition_period = YEAR
    unit = GBP


class charitable_investment_gifts(Variable):
    value_type = float
    entity = Person
    label = "Gifts of qualifying investment or property to charities"
    definition_period = YEAR
    unit = GBP


class other_deductions(Variable):
    value_type = float
    entity = Person
    label = "All other tax deductions"
    definition_period = YEAR
    unit = GBP


class allowances(Variable):
    value_type = float
    entity = Person
    label = "Allowances applicable to adjusted net income"
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        ALLOWANCES = [
            "personal_allowance",
            "blind_persons_allowance",
            "gift_aid",
            "covenanted_payments",
            "charitable_investment_gifts",
            "other_deductions",
        ]
        return add(person, period, ALLOWANCES)



from openfisca_uk.model_api import *


class fuel_duty(Variable):
    label = "Fuel duty (cars only)"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        fd = parameters(period).hmrc.fuel_duty
        petrol_litres = household("petrol_litres", period)
        diesel_litres = household("diesel_litres", period)
        return fd.petrol_and_diesel * (petrol_litres + diesel_litres)


class baseline_fuel_duty(Variable):
    label = "Baseline fuel duty (cars only)"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP


class change_in_fuel_duty(Variable):
    label = "Change in fuel duty"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    formula = change_over_baseline(fuel_duty)



from openfisca_uk.data import EnhancedFRS
from openfisca_uk import Microsimulation
from openfisca_tools.testing import generate_tests


test_parameter = generate_tests(
    Microsimulation(dataset=EnhancedFRS, year=2022)
)



from openfisca_uk.model_api import *


class sdlt_on_residential_property_transactions(Variable):
    label = "Stamp Duty on residential property"
    documentation = "Tax charge from purchase of residential property"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP
    reference = "https://www.legislation.gov.uk/ukpga/2003/14/section/55"

    def formula(household, period, parameters):
        stamp_duty = parameters(period).hmrc.stamp_duty
        # Tax on main-home purchases
        price_limit = stamp_duty.residential.purchase.main.first.max
        price = household("main_residential_property_purchased", period)
        residential_purchase_qualifies_as_first_buy = household(
            "main_residential_property_purchased_is_first_home", period
        ) & (price < price_limit)
        main_residential_purchase_tax = where(
            residential_purchase_qualifies_as_first_buy,
            stamp_duty.residential.purchase.main.first.rate.calc(price),
            stamp_duty.residential.purchase.main.subsequent.calc(price),
        )
        # Tax on second-home purchases
        second_home_price = household(
            "additional_residential_property_purchased", period
        )
        price = where(
            second_home_price < stamp_duty.residential.purchase.additional.min,
            0,
            second_home_price,
        )
        additional_residential_purchase_tax = (
            stamp_duty.residential.purchase.additional.rate.calc(price)
        )
        return (
            main_residential_purchase_tax + additional_residential_purchase_tax
        )


class sdlt_on_residential_property_rent(Variable):
    label = "Stamp Duty on residential property"
    documentation = "Tax charge from rental of residential property"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP
    reference = "https://www.legislation.gov.uk/ukpga/2003/14/schedule/5"

    def formula(household, period, parameters):
        stamp_duty = parameters(period).hmrc.stamp_duty
        cumulative_rent = household("cumulative_residential_rent", period)
        rent = household("rent", period)
        return stamp_duty.residential.rent.calc(
            cumulative_rent + rent
        ) - stamp_duty.residential.rent.calc(cumulative_rent)


class sdlt_on_non_residential_property_transactions(Variable):
    label = "Stamp Duty on non-residential property"
    documentation = "Tax charge from purchase of non-residential property"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP
    reference = "https://www.legislation.gov.uk/ukpga/2003/14/section/55"

    def formula(household, period, parameters):
        stamp_duty = parameters(period).hmrc.stamp_duty
        price = household("non_residential_property_purchased", period)
        return stamp_duty.non_residential.purchase.calc(price)


class sdlt_on_non_residential_property_rent(Variable):
    label = "Stamp Duty on non-residential property"
    documentation = "Tax charge from rental of non-residential property"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP
    reference = "https://www.legislation.gov.uk/ukpga/2003/14/schedule/5"

    def formula(household, period, parameters):
        stamp_duty = parameters(period).hmrc.stamp_duty
        cumulative_rent = household("cumulative_non_residential_rent", period)
        rent = household("non_residential_rent", period)
        return stamp_duty.non_residential.rent.calc(
            cumulative_rent + rent
        ) - stamp_duty.non_residential.rent.calc(cumulative_rent)


class sdlt_on_transactions(Variable):
    label = "SDLT on property transactions"
    documentation = "Stamp Duty Land Tax on property transfers"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP
    reference = "https://www.legislation.gov.uk/ukpga/2003/14/section/55"

    def formula(household, period):
        return household(
            "sdlt_on_residential_property_transactions", period
        ) + household("sdlt_on_non_residential_property_transactions", period)


class sdlt_on_rent(Variable):
    label = "SDLT on property rental"
    documentation = "Stamp Duty Land Tax on property rental agreements"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP
    reference = "https://www.legislation.gov.uk/ukpga/2003/14/section/55"

    def formula(household, period):
        return household(
            "sdlt_on_residential_property_rent", period
        ) + household("sdlt_on_non_residential_property_rent", period)


class sdlt_liable(Variable):
    label = "Liable for Stamp Duty"
    documentation = "Whether the household is liable for Stamp Duty Land Tax"
    entity = Household
    definition_period = YEAR
    value_type = bool
    unit = GBP

    def formula(household, period):
        country = household("country", period)
        countries = country.possible_values
        return np.isin(
            country.decode_to_str(),
            [countries.ENGLAND.name, countries.NORTHERN_IRELAND.name],
        )


class baseline_corporate_sdlt(Variable):
    label = "Stamp Duty (corporations, baseline)"
    documentation = "Stamp Duty paid by corporations, incident on this household in the baseline"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        sd = parameters(period).hmrc.stamp_duty.statistics
        return household("shareholding", period) * (
            sd.residential.corporate.revenue
            + sd.non_residential.corporate.revenue
        )


class corporate_sdlt(Variable):
    label = "Stamp Duty (corporations)"
    documentation = (
        "Stamp Duty paid by corporations, incident on this household"
    )
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    formula = baseline_corporate_sdlt.formula


class corporate_sdlt_change_incidence(Variable):
    label = "Change in corporate Stamp Duty (expected)"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    formula = change_over_baseline(corporate_sdlt)


class stamp_duty_land_tax(Variable):
    label = "Stamp Duty Land Tax"
    documentation = "Total tax liability for Stamp Duty Land Tax"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP
    reference = "https://www.legislation.gov.uk/ukpga/2003/14/part/4"

    def formula(household, period):
        return household("sdlt_liable", period) * add(
            household,
            period,
            [
                "sdlt_on_transactions",
                "sdlt_on_rent",
            ],
        )


class expected_sdlt(Variable):
    label = "Stamp Duty (expected)"
    documentation = "Expected value of Stamp Duty Land Tax"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period):
        return (
            household.state("property_sale_rate", period)
            * household("stamp_duty_land_tax", period)
        ) + household("corporate_sdlt", period)


class baseline_expected_sdlt(Variable):
    label = "Stamp Duty (expected, baseline)"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP


class change_in_expected_sdlt(Variable):
    label = "Change in expected Stamp Duty"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    formula = change_over_baseline(expected_sdlt)



from openfisca_uk.model_api import *


class unused_personal_allowance(Variable):
    value_type = float
    entity = Person
    label = "Unused personal allowance"
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        return max_(
            person("personal_allowance", period)
            - person("adjusted_net_income", period),
            0,
        )


class meets_marriage_allowance_income_conditions(Variable):
    label = "Meets Marriage Allowance income conditions"
    documentation = "Whether this person (and their partner) meets the conditions for this person to be eligible for the Marriage Allowance, as set out in the Income Tax Act 2007 sections 55B and 55C"
    entity = Person
    definition_period = YEAR
    value_type = bool
    reference = "https://www.legislation.gov.uk/ukpga/2007/3/section/55B"

    def formula(person, period):
        band = person("tax_band", period)
        bands = band.possible_values
        return (band != bands.HIGHER) & (band != bands.ADDITIONAL)


class partners_unused_personal_allowance(Variable):
    label = "Partner's unused personal allowance"
    documentation = "The personal tax allowance not used by this person's partner, if they exist"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(person, period, parameters):
        is_adult = person("is_adult", period)
        pa = person("unused_personal_allowance", period)
        return person.benunit.sum(is_adult * pa) - pa


class marriage_allowance(Variable):
    value_type = float
    entity = Person
    label = "Marriage Allowance for the year (a tax-reducer, rather than an allowance or tax relief)"
    definition_period = YEAR
    reference = "https://www.legislation.gov.uk/ukpga/2007/3/part/3/chapter/3A"
    unit = GBP

    def formula(person, period, parameters):
        marital = person("marital_status", period)
        married = marital == marital.possible_values.MARRIED
        eligible = married & person(
            "meets_marriage_allowance_income_conditions", period
        )
        transferable_amount = person(
            "partners_unused_personal_allowance", period
        )
        allowances = parameters(period).tax.income_tax.allowances
        capped_percentage = allowances.marriage_allowance.max
        max_amount = allowances.personal_allowance.amount * capped_percentage
        return eligible * min_(transferable_amount, max_amount)



from openfisca_uk.model_api import *


class armed_forces_independence_payment(Variable):
    label = "Armed Forces Independence Payment"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = GBP



import logging
from typing import Tuple
import pandas as pd
from openfisca_tools.data import Dataset, PrivateDataset
import h5py
from openfisca_uk.data.datasets.lcfs.raw_lcfs import RawLCFS
from openfisca_uk.data.storage import OPENFISCA_UK_MICRODATA_FOLDER

CATEGORY_NAMES = dict(
    # Top-level COICOP categories
    P601="Food and non-alcoholic beverages consumption",
    P602="Alcohol and tobacco consumption",
    P603="Clothing and footwear consumption",
    P604="Housing, water and electricity consumption",
    P605="Household furnishings consumption",
    P606="Health consumption",
    P607="Transport consumption",
    P608="Communication consumption",
    P609="Recreation consumption",
    P610="Education consumption",
    P611="Restaurants and hotels consumption",
    P612="Miscellaneous consumption",
    # Specific items
    C72211="Petrol spending",
    C72212="Diesel spending",
)

name_to_variable_name = {
    category: category.replace(",", "")
    .replace(" ", "_")
    .replace("-", "_")
    .lower()
    + ("_consumption" if category[1:] == "P" else "")
    for category in CATEGORY_NAMES.values()
}

CATEGORY_VARIABLES = list(name_to_variable_name.values())

HOUSEHOLD_LCF_RENAMES = {
    "G018": "is_adult",
    "G019": "is_child",
    "Gorx": "region",
}
PERSON_LCF_RENAMES = {
    "B303p": "employment_income",
    "B3262p": "self_employment_income",
    "B3381": "state_pension",
    "P049p": "pension_income",
}
REGIONS = {
    1: "NORTH_EAST",
    2: "NORTH_WEST",
    3: "YORKSHIRE",
    4: "EAST_MIDLANDS",
    5: "WEST_MIDLANDS",
    6: "EAST_OF_ENGLAND",
    7: "LONDON",
    8: "SOUTH_EAST",
    9: "SOUTH_WEST",
    10: "WALES",
    11: "SCOTLAND",
    12: "NORTHERN_IRELAND",
}


class LCFS(PrivateDataset):
    name = "lcfs"
    label = "LCFS"
    data_format = Dataset.ARRAYS
    folder_path = OPENFISCA_UK_MICRODATA_FOLDER

    is_openfisca_compatible = True

    def generate(self, year: int):
        if year in self.years:
            self.remove(year)
        # Load raw FRS tables
        year = int(year)

        if len(RawLCFS.years) == 0:
            raise FileNotFoundError(
                "Raw LCFS not found. Please run `openfisca-uk data lcfs generate [year]` first."
            )

        if year > max(RawLCFS.years):
            logging.warning("Uprating a previous version of the LCFS.")
            if len(self.years) == 0:
                self.generate(max(RawLCFS.years))
            if len(self.years) > 0:
                lcfs_year = max(self.years)
                from openfisca_uk import Microsimulation

                sim = Microsimulation(dataset=self, year=max(self.years))
                lcfs = h5py.File(self.file(year), mode="w")
                for variable in self.keys(lcfs_year):
                    lcfs[variable] = sim.calc(variable).values
                lcfs.close()
                return
        households = RawLCFS.load(2019, "lcfs_2019_dvhh_ukanon")
        people = RawLCFS.load(2019, "lcfs_2019_dvper_ukanon201920")
        spending = (
            households[list(CATEGORY_NAMES.keys())].unstack().reset_index()
        )
        spending.columns = "category", "household", "spending"
        spending["household"] = households.case[spending.household].values
        households = households.set_index("case")
        spending.category = spending.category.map(CATEGORY_NAMES).map(
            name_to_variable_name
        )
        spending.spending *= 52
        spending["weight"] = (
            households.weighta[spending.household].values * 100
        )
        spending = pd.DataFrame(spending)

        for category in spending.category.unique():
            spending[category] = (
                spending.category == category
            ) * spending.spending

        lcf_df = (
            pd.DataFrame(
                spending[["household", "weight"] + CATEGORY_VARIABLES]
            )
            .groupby("household")
            .sum()
        )

        # Add in LCFS variables that also appear in the FRS-based microsimulation model

        lcf_household_vars = households[
            list(HOUSEHOLD_LCF_RENAMES.keys())
        ].rename(columns=HOUSEHOLD_LCF_RENAMES)
        lcf_person_vars = (
            people[list(PERSON_LCF_RENAMES) + ["case"]]
            .rename(columns=PERSON_LCF_RENAMES)
            .groupby("case")
            .sum()
        )

        lcf_with_demographics = pd.concat(
            [
                lcf_df,
                lcf_household_vars,
                lcf_person_vars,
            ],
            axis=1,
        )

        # LCFS incomes are weekly - convert to annual
        for variable in PERSON_LCF_RENAMES.values():
            lcf_with_demographics[variable] *= 52

        lcf_with_demographics.region = lcf_with_demographics.region.map(
            REGIONS
        )
        lcfs = lcf_with_demographics.sort_index()

        lcfs = lcfs.rename(columns=dict(weight="household_weight"))

        entity_index = (
            lcfs.index.values
        )  # One-person households for simplicity for now

        with h5py.File(self.file(year), mode="w") as f:
            for entity_id_var in [
                "person_id",
                "benunit_id",
                "household_id",
                "person_benunit_id",
                "person_household_id",
            ]:
                f[entity_id_var] = entity_index

            f["person_benunit_role"] = ["adult"] * len(entity_index)
            f["person_household_role"] = ["adult"] * len(entity_index)
            f["person_state_id"] = [1] * len(entity_index)
            f["state_id"] = [1]

            for variable in lcfs.columns:
                f[variable] = lcfs[variable].values


LCFS = LCFS()



from openfisca_uk.model_api import *
from openfisca_uk.variables.household.demographic.locations import (
    LocalAuthority,
)
import pandas as pd

PC = pd.read_csv(DATA_FOLDER / "geography" / "postcodes.csv")
LA = pd.read_csv(DATA_FOLDER / "geography" / "local_authorities.csv")
LA_safe_name = (
    lambda x: x.upper()
    .replace(" ", "_")
    .replace("&", "")
    .replace("-", "_")
    .replace("__", "_")
    .replace("'", "")
    .replace(",", "")
    .replace(".", "")
)
LAC_LA = LA.set_index("LAD20CD").LAD20NM.apply(LA_safe_name).sort_values()
LAC_LA = LAC_LA[(PC.laua[PC.laua.isin(LAC_LA.index)]).unique()]
PC_LAC = PC.set_index("pcd").laua
LA_index = (
    LAC_LA.sort_values()
    .reset_index()
    .reset_index()
    .set_index("LAD20NM")["index"]
)


class postcode(Variable):
    value_type = str
    entity = Household
    label = "Postcode for the household"
    definition_period = YEAR


class local_authority(Variable):
    value_type = Enum
    possible_values = LocalAuthority
    default_value = LocalAuthority.MAIDSTONE
    entity = Household
    label = "The Local Authority for the household"
    definition_period = YEAR

    def formula(household, period, parameters):
        return LA_index[LAC_LA[PC_LAC[household("postcode", period)]]]


class with_postcode_features(Reform):
    def apply(self):
        self.add_variable(postcode)
        self.update_variable(local_authority)



# -*- coding: utf-8 -*-

import os
from openfisca_uk import entities
import os
from openfisca_core.taxbenefitsystems import TaxBenefitSystem

COUNTRY_DIR = os.path.dirname(os.path.abspath(__file__))


class CountryTaxBenefitSystem(TaxBenefitSystem):
    CURRENCY = "£"

    def __init__(self):
        # We initialize our tax and benefit system with the general constructor
        super(CountryTaxBenefitSystem, self).__init__(entities.entities)

        # We add to our tax and benefit system all the variables
        self.add_variables_from_directory(
            os.path.join(COUNTRY_DIR, "variables")
        )

        # We add to our tax and benefit system all the legislation parameters defined in the  parameters files
        param_path = os.path.join(COUNTRY_DIR, "parameters")
        self.load_parameters(param_path)

        self.parameters.baseline = self.parameters.clone()

        # We define which variable, parameter and simulation example will be used in the OpenAPI specification
        self.open_api_config = {
            "variable_example": "JSA",
            "parameter_example": "taxes.income_tax.income_tax",
            "simulation_example": None,
        }



from openfisca_uk.model_api import *


@uprated(by="uprating.september_cpi")
class ESA_contrib(Variable):
    value_type = float
    entity = Person
    label = "ESA (contribution-based)"
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        return person("ESA_contrib_reported", period)


class ESA_contrib_reported(Variable):
    value_type = float
    entity = Person
    label = "Employment and Support Allowance (contribution-based) (reported)"
    definition_period = YEAR
    unit = GBP



from openfisca_uk.model_api import *


@uprated(by="wealth.national_balance_sheet.household.net_worth")
class total_wealth(Variable):
    label = "Total wealth"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    formula = sum_of_variables(["property_wealth", "corporate_wealth"])



from openfisca_core.model_api import *
from openfisca_core.parameters.parameter_scale import ParameterScale
from openfisca_uk.entities import *
from openfisca_uk.tools.general import *
from datetime import datetime

DATE = datetime.now()
YEAR, MONTH, DAY = DATE.year, DATE.month, DATE.day
CURRENT_INSTANT = DATE.strftime("%Y-%m-%d")


def use_current_parameters(date: str = CURRENT_INSTANT) -> Reform:
    def modify_parameters(parameters: ParameterNode):
        for child in parameters.get_descendants():
            if isinstance(child, Parameter):
                current_value = child(date)
                child.update(period=f"year:{YEAR-10}:20", value=current_value)
            elif isinstance(child, ParameterScale):
                for bracket in child.brackets:
                    if "rate" in bracket.children:
                        current_rate = bracket.rate(date)
                        bracket.rate.update(
                            period=f"year:{YEAR-10}:20", value=current_rate
                        )
                    if "threshold" in bracket.children:
                        current_threshold = bracket.threshold(date)
                        bracket.threshold.update(
                            period=f"year:{YEAR-10}:20",
                            value=current_threshold,
                        )
        return parameters

    class reform(Reform):
        def apply(self):
            self.modify_parameters(modify_parameters)

    return reform



def single_person_UC(sim):
    sim.add_person(
        name="adult", age=26, is_household_head=True, is_benunit_head=True
    )
    sim.add_benunit(adults=["adult"], would_claim_UC=True)
    sim.add_household(adults=["adult"])
    return sim


def couple_UC(sim):
    sim.add_person(
        name="adult", age=26, is_household_head=True, is_benunit_head=True
    )
    sim.add_person(name="adult_2", age=27)
    sim.add_benunit(adults=["adult", "adult_2"], would_claim_UC=True)
    sim.add_household(adults=["adult", "adult_2"])
    return sim


def single_person_UC_one_child(sim):
    sim.add_person(
        name="adult", age=26, is_household_head=True, is_benunit_head=True
    )
    sim.add_person(name="child", age=4)
    sim.add_benunit(adults=["adult"], children=["child"], would_claim_UC=True)
    sim.add_household(adults=["adult"], children=["child"])
    return sim


def couple_UC_one_child(sim):
    sim.add_person(
        name="adult", age=26, is_household_head=True, is_benunit_head=True
    )
    sim.add_person(name="adult_2", age=27)
    sim.add_person(name="child", age=4)
    sim.add_benunit(
        adults=["adult", "adult_2"], children=["child"], would_claim_UC=True
    )
    sim.add_household(adults=["adult", "adult_2"], children=["child"])
    return sim


def single_person_UC_two_children(sim):
    sim.add_person(
        name="adult", age=26, is_household_head=True, is_benunit_head=True
    )
    sim.add_person(name="child", age=4)
    sim.add_person(name="child_2", age=6)
    sim.add_benunit(
        adults=["adult"],
        children=["child", "child_2"],
        would_claim_UC=True,
        claims_CB=True,
    )
    sim.add_household(adults=["adult"], children=["child", "child_2"])
    return sim


def couple_UC_two_children(sim):
    sim.add_person(
        name="adult", age=26, is_household_head=True, is_benunit_head=True
    )
    sim.add_person(name="adult_2", age=27)
    sim.add_person(name="child", age=4)
    sim.add_person(name="child_2", age=6)
    sim.add_benunit(
        adults=["adult", "adult_2"],
        children=["child", "child_2"],
        would_claim_UC=True,
    )
    sim.add_household(
        adults=["adult", "adult_2"], children=["child", "child_2"]
    )
    return sim



from openfisca_uk.model_api import *


class is_guarantee_credit_eligible(Variable):
    label = "Guarantee Credit eligible"
    entity = BenUnit
    definition_period = YEAR
    value_type = bool
    unit = GBP

    def formula(benunit, period, parameters):
        income = benunit("pension_credit_income", period)
        minimum_guarantee = benunit("minimum_guarantee", period)
        return income < minimum_guarantee



import re
import shutil
from openfisca_tools.data import PrivateDataset
import pandas as pd
from pathlib import Path
from tqdm import tqdm
from openfisca_uk.data.storage import OPENFISCA_UK_MICRODATA_FOLDER


class RawFRS(PrivateDataset):
    name = "raw_frs"
    label = "Raw FRS"
    folder_path = OPENFISCA_UK_MICRODATA_FOLDER
    is_openfisca_compatible = False

    filename_by_year = {
        2019: "raw_frs_2019.h5",
    }

    def generate(self, year: int, ukds_tab_zipfile: str):
        """Generates the raw FRS tabular dataset from the TAB zip archive
        downloadable from the UKDS.

        Args:
            year (int): The year of the FRS to generate.
            ukds_tab_zipfile (str): The path to the TAB zip archive.
        """

        folder = Path(ukds_tab_zipfile)
        year = str(year)

        if not folder.exists():
            raise FileNotFoundError("Invalid path supplied.")

        new_folder = self.folder_path / "tmp"
        shutil.unpack_archive(folder, new_folder)
        folder = new_folder

        main_folder = next(folder.iterdir())
        tab_folder = main_folder / "tab"
        if tab_folder.exists():
            criterion = re.compile("[a-z]+\.tab")
            data_files = [
                path
                for path in tab_folder.iterdir()
                if criterion.match(path.name)
            ]
            task = tqdm(data_files, desc="Saving raw data tables")
            with pd.HDFStore(RawFRS.file(year)) as file:
                for filepath in task:
                    task.set_description(
                        f"Saving raw data tables ({filepath.name})"
                    )
                    table_name = filepath.name.replace(".tab", "")
                    df = pd.read_csv(
                        filepath, delimiter="\t", low_memory=False
                    ).apply(pd.to_numeric, errors="coerce")
                    df.columns = df.columns.str.upper()
                    if "PERSON" in df.columns:
                        df["person_id"] = (
                            df.SERNUM * 1e2 + df.BENUNIT * 1e1 + df.PERSON
                        ).astype(int)
                    if "BENUNIT" in df.columns:
                        df["benunit_id"] = (
                            df.SERNUM * 1e2 + df.BENUNIT * 1e1
                        ).astype(int)
                    if "SERNUM" in df.columns:
                        df["household_id"] = (df.SERNUM * 1e2).astype(int)
                    if table_name in ("adult", "child"):
                        df.set_index("person_id", inplace=True)
                    elif table_name == "benunit":
                        df.set_index("benunit_id", inplace=True)
                    elif table_name == "househol":
                        df.set_index("household_id", inplace=True)
                    file[table_name] = df
        else:
            raise FileNotFoundError("Could not find the TAB files.")

        tmp_folder = self.folder_path / "tmp"
        if tmp_folder.exists():
            shutil.rmtree(tmp_folder)


RawFRS = RawFRS()



import logging
from openfisca_tools.data import PrivateDataset, Dataset
from openfisca_uk.data.storage import OPENFISCA_UK_MICRODATA_FOLDER
from ..extension import ExtendedFRS
import h5py


class CalibratedFRS(PrivateDataset):
    name = "calibrated_frs"
    label = "Calibrated FRS"
    data_format = Dataset.TIME_PERIOD_ARRAYS

    folder_path = OPENFISCA_UK_MICRODATA_FOLDER

    def generate(self, year: int):
        if year not in ExtendedFRS.years:
            ok = input(
                f"Extended FRS not found for year {year}. Generate it? (y/n)"
            )
            if ok == "y":
                ExtendedFRS.generate(year)

        # Copy all data from the extended FRS

        extended_frs = ExtendedFRS.load(year)
        calibrated_frs = h5py.File(self.file(year), mode="w")
        for variable in extended_frs.keys():
            for period in extended_frs[variable].keys():
                calibrated_frs[f"{variable}/{period}"] = extended_frs[
                    variable
                ][period][...]
        extended_frs.close()
        calibrated_frs.close()

        from .calibrate import HouseholdWeights

        weights = HouseholdWeights(
            start_year=2022,
            end_year=2027,
        )
        weights.calibrate(
            validation_split=0,
            num_epochs=300,
            learning_rate=1e2,
            dataset=self,
        )

        weights.save()


CalibratedFRS = CalibratedFRS()



from openfisca_uk.model_api import *

"""
The calculation of income tax is specified by the Income Tax Act 2007 s. 23
which outlines the main steps. Step 1 indicates that the components of 'total income'
should be added together, but the Act itself does not specify what those components are.
Instead, other Acts (the Income Tax (Earnings and Pensions) Act 2003 and the Income Tax 
(Trading and Other Income) Act 2005) impose the charges to income tax on each component of
income. The relevant sections of each act imposing charges to income tax are given 
for each component of income below. The components are guided by the article at 
https://www.pruadviser.co.uk/knowledge-literature/knowledge-library/the-seven-steps-required-to-calculate-an-individuals-income-tax-liability/
"""


@uprated(by="uprating.earnings")
class employment_income(Variable):
    value_type = float
    entity = Person
    label = "Employment income"
    documentation = "Total income from employment"
    definition_period = YEAR
    unit = GBP
    reference = "Income Tax (Earnings and Pensions) Act 2003 s. 1(1)(a)"
    quantity_type = FLOW


@uprated(by="uprating.earnings")
class pension_income(Variable):
    value_type = float
    entity = Person
    label = "Pension income"
    documentation = "Income from private or occupational pensions (not including State Pension)"
    definition_period = YEAR
    unit = GBP
    reference = "Income Tax (Earnings and Pensions) Act 2003 s. 1(1)(b)"
    quantity_type = FLOW


class total_pension_income(Variable):
    label = "Total pension income"
    documentation = "Private, personal and State Pension income"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = GBP

    formula = sum_of_variables(["pension_income", "state_pension"])


class social_security_income(Variable):
    value_type = float
    entity = Person
    label = "Income from social security for tax purposes"
    definition_period = YEAR
    reference = "Income Tax (Earnings and Pensions) Act 2003 s. 1(1)(b)"
    unit = GBP

    def formula(person, period, parameters):
        COMPONENTS = [
            "state_pension",
            "incapacity_benefit",
            "JSA_contrib",
            "ESA_contrib",
            "carers_allowance",
        ]
        return add(person, period, COMPONENTS, options=[ADD])


@uprated(by="uprating.earnings")
class self_employment_income(Variable):
    value_type = float
    entity = Person
    label = "Self-employment income"
    documentation = "Income from self-employment profits"
    definition_period = YEAR
    unit = GBP
    reference = "Income Tax (Trading and Other Income) Act 2005 s. 1(1)(a)"
    quantity_type = FLOW


@uprated(by="uprating.earnings")
class property_income(Variable):
    value_type = float
    entity = Person
    label = "Rental income"
    documentation = "Income from rental of property"
    definition_period = YEAR
    reference = "Income Tax (Trading and Other Income) Act 2005 s. 1(1)(b)"
    unit = GBP
    quantity_type = FLOW


@uprated(by="uprating.earnings")
class savings_interest_income(Variable):
    value_type = float
    entity = Person
    label = "Savings interest income"
    documentation = "Income from interest on savings, gross of tax"
    definition_period = YEAR
    reference = "Income Tax (Trading and Other Income) Act 2005 s. 365(1)(a)"
    unit = GBP
    quantity_type = FLOW


@uprated(by="uprating.earnings")
class dividend_income(Variable):
    value_type = float
    entity = Person
    label = "Income from dividends"
    documentation = "Total income from dividends, gross of tax"
    definition_period = YEAR
    reference = "Income Tax (Trading and Other Income) Act 2005 s. 365(1)(b-d)"
    unit = GBP
    quantity_type = FLOW



from openfisca_uk.model_api import *
from openfisca_uk.variables.household.demographic.household import TenureType


class legacy_benefits(Variable):
    label = "Legacy benefits"
    entity = BenUnit
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(benunit, period, parameters):
        BENEFITS = [
            "JSA_income",
            "housing_benefit",
            "income_support",
            "ESA_income",
            "working_tax_credit",
            "child_tax_credit",
        ]
        return add(benunit, period, BENEFITS)


class would_claim_UC(Variable):
    value_type = bool
    entity = BenUnit
    label = "Would claim Universal Credit"
    documentation = (
        "Whether this family would claim Universal Credit if eligible"
    )
    definition_period = YEAR

    def formula(benunit, period, parameters):
        on_legacy_benefits = benunit("claims_legacy_benefits", period)
        claims_all_entitled_benefits = benunit(
            "claims_all_entitled_benefits", period
        )
        current_uc_claimant = (
            aggr(benunit, period, ["universal_credit_reported"]) > 0
        )
        baseline_uc = benunit("baseline_has_universal_credit", period)
        takeup_rate = parameters(period).benefit.universal_credit.takeup
        return select(
            [
                current_uc_claimant
                | (claims_all_entitled_benefits & ~on_legacy_benefits),
                (~baseline_uc & ~on_legacy_benefits),
                True,
            ],
            [
                True,  # Claims Universal Credit in the baseline
                random(benunit)
                < takeup_rate,  # New UC eligibility from a reform
                False,  # Always non-claimant
            ],
        )


class is_UC_eligible(Variable):
    value_type = bool
    entity = BenUnit
    label = "Universal Credit eligible"
    documentation = "Whether this family is eligible for Universal Credit"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        return benunit.any(benunit.members("is_WA_adult", period))


class universal_credit_reported(Variable):
    value_type = float
    entity = Person
    label = "Universal Credit (reported)"
    documentation = "Reported amount of Universal Credit"
    definition_period = YEAR
    unit = GBP


class UC_maximum_amount(Variable):
    value_type = float
    entity = BenUnit
    label = "Maximum Universal Credit amount"
    definition_period = YEAR
    unit = GBP

    def formula(benunit, period):
        ELEMENTS = [
            "UC_standard_allowance",
            "UC_child_element",
            "UC_disability_elements",
            "UC_carer_element",
            "UC_housing_costs_element",
            "UC_childcare_element",
        ]
        return add(benunit, period, ELEMENTS)


class UCClaimantType(Enum):
    SINGLE_YOUNG = "Single, under 25"
    SINGLE_OLD = "Single, 25 or over"
    COUPLE_YOUNG = "Couple, both under 25"
    COUPLE_OLD = "Couple, one over 25"


class UC_claimant_type(Variable):
    value_type = Enum
    possible_values = UCClaimantType
    default_value = UCClaimantType.SINGLE_YOUNG
    entity = BenUnit
    label = "UC claimant type"
    documentation = (
        "The category of the UC claimant, assuming their eligibilty to UC"
    )
    definition_period = YEAR

    def formula(benunit, period):
        is_single = benunit("is_single", period)
        any_over_25 = benunit("eldest_adult_age", period.this_year) >= 25
        return select(
            [
                is_single & ~any_over_25,
                is_single & any_over_25,
                ~is_single & ~any_over_25,
                ~is_single & any_over_25,
            ],
            [
                UCClaimantType.SINGLE_YOUNG,
                UCClaimantType.SINGLE_OLD,
                UCClaimantType.COUPLE_YOUNG,
                UCClaimantType.COUPLE_OLD,
            ],
        )


class UC_standard_allowance(Variable):
    value_type = float
    entity = BenUnit
    label = "Universal Credit standard allowance"
    definition_period = YEAR
    unit = GBP

    def formula(benunit, period, parameters):
        SA = parameters(period).benefit.universal_credit.standard_allowance
        return SA.amount[benunit("UC_claimant_type", period)] * MONTHS_IN_YEAR


class is_child_born_before_child_limit(Variable):
    value_type = bool
    entity = Person
    label = "Born before child limit (exempt)"
    definition_period = YEAR

    def formula(person, period, parameters):
        UC = parameters(period).benefit.universal_credit
        start_year = UC.elements.child.limit.start_year
        born_before_limit = person("birth_year", period) < start_year
        return person("is_child", period) & born_before_limit


class UC_individual_child_element(Variable):
    value_type = float
    entity = Person
    label = "Universal Credit child element"
    documentation = "For this child, given Universal Credit eligibility"
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        UC = parameters(period).benefit.universal_credit
        child_index = person("child_index", period)
        born_before_limit = person("is_child_born_before_child_limit", period)
        child_limit_applying = where(
            ~born_before_limit, UC.elements.child.limit.child_count, inf
        )
        is_eligible = (child_index != -1) & (
            child_index <= child_limit_applying
        )
        return (
            select(
                [
                    (child_index == 1) & born_before_limit & is_eligible,
                    child_index == 1 & is_eligible,
                    (child_index > 1) & is_eligible,
                    True,
                ],
                [
                    UC.elements.child.first.higher_amount,
                    UC.elements.child.amount,
                    UC.elements.child.amount,
                    0,
                ],
            )
            * MONTHS_IN_YEAR
        )


class num_UC_eligible_children(Variable):
    value_type = int
    entity = BenUnit
    label = "Children eligible for Universal Credit"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        children_born_before_limit = aggr(
            benunit, period, ["is_child_born_before_child_limit"]
        )
        child_limit = parameters(
            period
        ).benefit.universal_credit.elements.child.limit.child_count
        spaces_left = clip(
            child_limit - children_born_before_limit, 0, child_limit
        )
        spaces_filled = min_(
            spaces_left,
            benunit("num_children", period) - children_born_before_limit,
        )
        return children_born_before_limit + spaces_filled


class UC_child_element(Variable):
    value_type = float
    entity = BenUnit
    label = "Universal Credit child element"
    definition_period = YEAR
    unit = GBP

    def formula(benunit, period, parameters):
        return aggr(benunit, period, ["UC_individual_child_element"])


class UC_carer_element(Variable):
    value_type = float
    entity = BenUnit
    label = "Universal Credit carer element"
    definition_period = YEAR
    unit = GBP

    def formula(benunit, period, parameters):
        UC = parameters(period).benefit.universal_credit
        has_carer = benunit("benunit_has_carer", period)
        return UC.elements.carer.amount * has_carer * MONTHS_IN_YEAR


class UC_housing_costs_element(Variable):
    value_type = float
    entity = BenUnit
    label = "Universal Credit housing costs element"
    definition_period = YEAR
    unit = GBP

    def formula(benunit, period, parameters):
        tenure_type = benunit.value_from_first_person(
            benunit.members.household("tenure_type", period).decode_to_str()
        )
        rent = benunit("benunit_rent", period)
        max_housing_costs = select(
            [
                np.isin(
                    tenure_type,
                    [
                        TenureType.RENT_FROM_COUNCIL.name,
                        tenure_type == TenureType.RENT_FROM_HA.name,
                    ],
                ),
                tenure_type == TenureType.RENT_PRIVATELY.name,
                True,
            ],
            [
                rent,
                min_(benunit("LHA_cap", period), rent),
                0,
            ],
        )
        return max_housing_costs - benunit("UC_non_dep_deductions", period)


class UC_non_dep_deduction_exempt(Variable):
    value_type = bool
    entity = Person
    label = "Not expected to contribute to housing costs for Universal Credit"
    definition_period = YEAR

    def formula(person, period, parameters):
        return (
            (person.benunit("pension_credit", period) > 0)
            | person("dla_sc_middle_plus", period)
            | (person("pip_dl", period) > 0)
            | (person("attendance_allowance", period) > 0)
            | person("receives_carers_allowance", period)
        )


class UC_individual_non_dep_deduction(Variable):
    value_type = float
    entity = Person
    label = "Universal Credit non-dependent deduction (individual)"
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        not_rent_liable = person.benunit("benunit_rent", period) == 0
        over_21 = person("age", period) >= 21
        exempt = person("UC_non_dep_deduction_exempt", period)
        deduction = parameters(
            period
        ).benefit.universal_credit.elements.housing.non_dep_deduction
        return deduction * ~exempt * over_21 * not_rent_liable * MONTHS_IN_YEAR


class UC_non_dep_deductions(Variable):
    value_type = float
    entity = BenUnit
    label = "Universal Credit non-dependent deductions"
    definition_period = YEAR
    unit = GBP

    def formula(benunit, period, parameters):
        # Deductions are made for non-dependents outside the benefit unit,
        # but within the household, who meet certain conditions. To do this,
        # we first calculate the non-dependent deduction for each person (from
        # the perspective of a different benefit unit). Then, to calculate
        # the deduction for non-dependents outside the benefit unit, we subtract
        # the total non-dependent deductions for the benefit unit members from
        # the deductions for household members.
        deductions = benunit.members("UC_individual_non_dep_deduction", period)
        non_dep_deductions_in_hh = benunit.max(
            benunit.members.household.sum(deductions)
        )
        non_dep_deductions_in_bu = benunit.sum(deductions)
        return non_dep_deductions_in_hh - non_dep_deductions_in_bu


class limited_capability_for_WRA(Variable):
    value_type = bool
    entity = Person
    label = "Assessed to have limited capability for work-related activity"
    documentation = """Whether this person has been assessed by the DWP as having limited capability for work or work-related activity"""
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("is_disabled_for_benefits", period)


class UC_LCWRA_element(Variable):
    value_type = float
    entity = BenUnit
    label = (
        "Universal Credit limited capability for work-related-activity element"
    )
    definition_period = YEAR
    unit = GBP

    def formula(benunit, period, parameters):
        UC = parameters(period).benefit.universal_credit
        person_amounts = (
            benunit.members("limited_capability_for_WRA", period)
            * UC.elements.disabled.amount
        )
        return benunit.sum(person_amounts) * MONTHS_IN_YEAR


class UC_individual_disabled_child_element(Variable):
    value_type = float
    entity = Person
    label = "Universal Credit disabled child element"
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        UC = parameters(period).benefit.universal_credit
        return (
            person("is_disabled_for_benefits", period)
            * person("is_child", period)
            * UC.elements.child.disabled.amount
        ) * MONTHS_IN_YEAR


class UC_individual_severely_disabled_child_element(Variable):
    value_type = float
    entity = Person
    label = "Universal Credit severely disabled child element"
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        UC = parameters(period).benefit.universal_credit
        return (
            person("is_severely_disabled_for_benefits", period)
            * person("is_child", period)
            * UC.elements.child.severely_disabled.amount
        ) * MONTHS_IN_YEAR


class UC_disability_elements(Variable):
    value_type = float
    entity = BenUnit
    label = "Universal Credit disability elements"
    definition_period = YEAR
    unit = GBP

    def formula(benunit, period):
        PERSONAL_ELEMENTS = [
            "UC_individual_disabled_child_element",
            "UC_individual_severely_disabled_child_element",
        ]
        personal_elements = aggr(benunit, period, PERSONAL_ELEMENTS)
        benunit_elements = benunit("UC_LCWRA_element", period)
        return personal_elements + benunit_elements


class UC_childcare_work_condition(Variable):
    value_type = bool
    entity = BenUnit
    label = "Meets Universal Credit childcare work condition"
    definition_period = YEAR
    reference = (
        "https://www.legislation.gov.uk/uksi/2013/376/regulation/32/2020-04-06"
    )

    def formula(benunit, period, parameters):
        person = benunit.members
        adult = person("is_adult", period)
        in_work = person("in_work", period)
        adults_not_in_work = adult & ~in_work
        # Benefit unit must not have any adults not in work.
        return ~benunit.any(adults_not_in_work)


class num_UC_eligible_children(Variable):
    value_type = int
    entity = BenUnit
    label = "Children eligible for Universal Credit"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        eligible_child = (
            benunit.members("UC_individual_child_element", period) > 0
        )
        return benunit.sum(eligible_child)


class UC_maximum_childcare(Variable):
    value_type = float
    entity = BenUnit
    label = "Maximum Universal Credit childcare element"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        UC = parameters(period).benefit.universal_credit
        children = benunit("num_UC_eligible_children", period)
        return (
            UC.elements.childcare.maximum[clip(children, 1, 2)]
            * MONTHS_IN_YEAR
        )


class UC_childcare_element(Variable):
    value_type = float
    entity = BenUnit
    label = "Universal Credit childcare element"
    definition_period = YEAR
    unit = GBP

    def formula(benunit, period, parameters):
        UC = parameters(period).benefit.universal_credit
        eligible_childcare_expenses = aggr(
            benunit, period, ["childcare_expenses"]
        )
        covered_expenses = (
            eligible_childcare_expenses * UC.elements.childcare.coverage_rate
        )
        childcare_element = min_(
            benunit("UC_maximum_childcare", period), covered_expenses
        )
        work_condition = benunit("UC_childcare_work_condition", period)
        return work_condition * childcare_element


class is_UC_work_allowance_eligible(Variable):
    value_type = bool
    entity = BenUnit
    label = "Family receives a Universal Credit Work Allowance"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        has_LCWRA = benunit.any(
            benunit.members("limited_capability_for_WRA", period)
        )
        has_children = benunit.any(benunit.members("is_child", period))
        return has_LCWRA | has_children


class UC_work_allowance(Variable):
    value_type = float
    entity = BenUnit
    label = "Universal Credit work allowance"
    definition_period = YEAR
    unit = GBP

    def formula(benunit, period, parameters):
        UC = parameters(period).benefit.universal_credit
        housing = benunit("UC_housing_costs_element", period)
        monthly_allowance = where(
            housing > 0,
            UC.means_test.work_allowance_with_housing,
            UC.means_test.work_allowance_without_housing,
        )
        eligible = benunit("is_UC_work_allowance_eligible", period)
        return eligible * monthly_allowance * MONTHS_IN_YEAR


class UC_earned_income(Variable):
    value_type = float
    entity = BenUnit
    label = "Universal Credit earned income (after disregards and tax)"
    definition_period = YEAR
    unit = GBP

    def formula(benunit, period, parameters):
        personal_gross_earned_income = aggr(
            benunit, period, ["UC_MIF_capped_earned_income"]
        )
        benunit_disregards = add(
            benunit, period, ["UC_work_allowance", "benunit_tax"]
        )
        person_disregards = aggr(benunit, period, ["pension_contributions"])
        disregards = benunit_disregards + person_disregards
        return max_(0, personal_gross_earned_income - disregards)


class UC_unearned_income(Variable):
    value_type = float
    entity = BenUnit
    label = "Universal Credit unearned income"
    definition_period = YEAR
    unit = GBP

    def formula(benunit, period, parameters):
        SOURCES = [
            "carers_allowance",
            "JSA_contrib",
            "pension_income",
            "savings_interest_income",
            "dividend_income",
        ]
        return aggr(benunit, period, SOURCES)


class UC_income_reduction(Variable):
    value_type = float
    entity = BenUnit
    label = "Reduction from income for Universal Credit"
    definition_period = YEAR
    unit = GBP

    def formula(benunit, period, parameters):
        UC = parameters(period).benefit.universal_credit
        earned_income_reduction = UC.means_test.reduction_rate * benunit(
            "UC_earned_income", period
        )
        unearned_income_reduction = benunit("UC_unearned_income", period)
        return max_(0, earned_income_reduction + unearned_income_reduction)


class universal_credit(Variable):
    value_type = float
    entity = BenUnit
    label = "Universal Credit"
    definition_period = YEAR
    unit = GBP

    def formula(benunit, period, parameters):
        amount = max_(
            0,
            benunit("UC_maximum_amount", period)
            - benunit("UC_income_reduction", period),
        )
        return (
            amount
            * benunit("would_claim_UC", period)
            * benunit("is_UC_eligible", period)
        )


class UC_MIF_applies(Variable):
    value_type = bool
    entity = Person
    label = "Minimum Income Floor applies"
    documentation = "Whether the Minimum Income Floor should be used to determine UC entitlement"
    reference = (
        "https://www.legislation.gov.uk/uksi/2013/376/regulation/62/2021-04-06"
    )
    definition_period = YEAR

    def formula(person, period, parameters):
        has_profits = person("self_employment_income", period) > 0
        in_startup_period = person("is_in_startup_period", period)
        return has_profits & ~in_startup_period


class is_in_startup_period(Variable):
    value_type = bool
    entity = Person
    label = "In a start-up period"
    documentation = (
        "Whether this person is in a 'start-up' period for Universal Credit"
    )
    definition_period = YEAR
    default_value = False


class UC_minimum_income_floor(Variable):
    value_type = float
    entity = Person
    label = "Minimum Income Floor"
    definition_period = YEAR
    unit = GBP
    reference = ""

    def formula(person, period, parameters):
        expected_hours = parameters(
            period
        ).benefit.universal_credit.work_requirements.default_expected_hours
        return person("minimum_wage", period) * expected_hours * WEEKS_IN_YEAR


class UC_MIF_capped_earned_income(Variable):
    value_type = float
    entity = Person
    label = "Universal Credit gross earned income (incl. MIF)"
    documentation = (
        "Gross earned income for UC, with MIF applied where applicable"
    )
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        INCOME_COMPONENTS = [
            "employment_income",
            "self_employment_income",
            "miscellaneous_income",
        ]
        bi = parameters(period).contrib.ubi_center.basic_income
        if bi.interactions.include_in_means_tests:
            INCOME_COMPONENTS.append("basic_income")
        personal_gross_earned_income = add(person, period, INCOME_COMPONENTS)
        floor = where(
            person("UC_MIF_applies", period),
            person("UC_minimum_income_floor", period),
            -inf,
        )
        return max_(personal_gross_earned_income, floor)


class baseline_universal_credit(Variable):
    label = "Universal Credit (baseline)"
    entity = BenUnit
    definition_period = YEAR
    value_type = float
    unit = GBP


class baseline_has_universal_credit(Variable):
    label = "Receives Universal Credit (baseline)"
    entity = BenUnit
    definition_period = YEAR
    value_type = bool
    default_value = True

    formula = baseline_is_nonzero(universal_credit)



from openfisca_uk.model_api import *
from openfisca_uk.variables.gov.dwp.pension_credit.pension_credit import (
    pension_credit,
)


class baseline_pension_credit(Variable):
    label = "Pension Credit (baseline)"
    entity = BenUnit
    definition_period = YEAR
    value_type = float
    unit = GBP


class baseline_has_pension_credit(Variable):
    label = "Receives Pension Credit (baseline)"
    entity = BenUnit
    definition_period = YEAR
    value_type = bool
    default_value = True

    formula = baseline_is_nonzero(pension_credit)


class would_claim_pc(Variable):
    label = "Would claim Pension Credit"
    entity = BenUnit
    definition_period = YEAR
    value_type = bool

    def formula(benunit, period, parameters):
        reported_pc = add(benunit, period, ["pension_credit_reported"]) > 0
        claims_all_entitled_benefits = benunit(
            "claims_all_entitled_benefits", period
        )
        baseline = benunit("baseline_has_pension_credit", period)
        takeup_rate = parameters(period).dwp.pension_credit.takeup
        return select(
            [reported_pc | claims_all_entitled_benefits, ~baseline, True],
            [
                True,
                random(benunit) < takeup_rate,
                False,
            ],
        )



from openfisca_uk.model_api import *


class weekly_NI_class_2(Variable):
    value_type = float
    entity = Person
    label = "Class 2 Contributions for National Insurance"
    definition_period = YEAR
    reference = "Social Security and Benefits Act 1992 s. 11"
    unit = GBP

    def formula(person, period, parameters):
        class_2 = parameters(period).tax.national_insurance.class_2
        profits = person("self_employment_income", period)
        over_threshold = profits >= class_2.small_profits_threshold
        return over_threshold * class_2.flat_rate * WEEKS_IN_YEAR


class NI_class_2(Variable):
    value_type = float
    entity = Person
    label = "Class 2 Contributions for National Insurance for the year"
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        return person("weekly_NI_class_2", period)



from openfisca_uk.model_api import *
from openfisca_uk.variables.gov.dwp.pip.pip import PIPCategory


class PIP_DL_reported(Variable):
    value_type = float
    entity = Person
    label = "PIP (self-care) (reported)"
    definition_period = YEAR
    unit = GBP


class pip_dl_category(Variable):
    label = "PIP (daily living) category"
    entity = Person
    definition_period = YEAR
    value_type = Enum
    possible_values = PIPCategory
    default_value = PIPCategory.NONE

    def formula(person, period, parameters):
        pip_dl = parameters(period).dwp.pip.daily_living
        SAFETY_MARGIN = 0.1  # Survey reported values could be slightly below eligible values when they should be above due to data manipulation
        reported_weekly_pip_dl = (
            person("PIP_DL_reported", period) / WEEKS_IN_YEAR
        )
        return select(
            [
                reported_weekly_pip_dl
                >= pip_dl.enhanced * (1 - SAFETY_MARGIN),
                reported_weekly_pip_dl
                >= pip_dl.standard * (1 - SAFETY_MARGIN),
                True,
            ],
            [
                PIPCategory.ENHANCED,
                PIPCategory.STANDARD,
                PIPCategory.NONE,
            ],
        )


class pip_dl(Variable):
    label = "PIP (daily living)"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(person, period, parameters):
        pip_dl = parameters(period).dwp.pip.daily_living
        category = person("pip_dl_category", period)
        return (
            select(
                [
                    category == PIPCategory.ENHANCED,
                    category == PIPCategory.STANDARD,
                    category == PIPCategory.NONE,
                ],
                [
                    pip_dl.enhanced,
                    pip_dl.standard,
                    0,
                ],
            )
            * WEEKS_IN_YEAR
        )


class receives_enhanced_pip_dl(Variable):
    label = "Receives enhanced PIP (daily living)"
    entity = Person
    definition_period = YEAR
    value_type = bool

    def formula(person, period, parameters):
        return person("pip_dl_category", period) == PIPCategory.ENHANCED



from openfisca_uk.model_api import *
from openfisca_uk.variables.misc.categories.lower_middle_or_higher import (
    LowerMiddleOrHigher,
)


class DLA_SC_reported(Variable):
    value_type = float
    entity = Person
    label = "DLA (self-care) (reported)"
    definition_period = YEAR
    unit = GBP


class dla_sc_category(Variable):
    label = "DLA (Self-care) category"
    entity = Person
    definition_period = YEAR
    value_type = Enum
    possible_values = LowerMiddleOrHigher
    default_value = LowerMiddleOrHigher.NONE

    def formula(person, period, parameters):
        dla_sc = parameters(period).dwp.dla.self_care
        SAFETY_MARGIN = 0.1  # Survey reported values could be slightly below eligible values when they should be above due to data manipulation
        reported_weekly_dla_sc = (
            person("DLA_SC_reported", period) / WEEKS_IN_YEAR
        )
        return select(
            [
                reported_weekly_dla_sc >= dla_sc.higher * (1 - SAFETY_MARGIN),
                reported_weekly_dla_sc >= dla_sc.middle * (1 - SAFETY_MARGIN),
                reported_weekly_dla_sc >= dla_sc.lower * (1 - SAFETY_MARGIN),
                True,
            ],
            [
                LowerMiddleOrHigher.HIGHER,
                LowerMiddleOrHigher.MIDDLE,
                LowerMiddleOrHigher.LOWER,
                LowerMiddleOrHigher.NONE,
            ],
        )


class dla_sc(Variable):
    label = "DLA (self-care)"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(person, period, parameters):
        dla_sc = parameters(period).dwp.dla.self_care
        category = person("dla_sc_category", period)
        return (
            select(
                [
                    category == LowerMiddleOrHigher.HIGHER,
                    category == LowerMiddleOrHigher.MIDDLE,
                    category == LowerMiddleOrHigher.LOWER,
                    category == LowerMiddleOrHigher.NONE,
                ],
                [
                    dla_sc.higher,
                    dla_sc.middle,
                    dla_sc.lower,
                    0,
                ],
            )
            * WEEKS_IN_YEAR
        )


class dla_sc_middle_plus(Variable):
    value_type = bool
    entity = Person
    label = "Receives at least middle-rate DLA (self-care)"
    definition_period = YEAR

    def formula(person, period, parameters):
        return is_in(
            person("dla_sc_category", period),
            [LowerMiddleOrHigher.MIDDLE, LowerMiddleOrHigher.HIGHER],
        )


class receives_highest_dla_sc(Variable):
    label = "Receives the highest DLA (self-care) category"
    entity = Person
    definition_period = YEAR
    value_type = bool

    def formula(person, period, parameters):
        return person("dla_sc_category", period) == LowerMiddleOrHigher.HIGHER



import numpy as np
from ...loss_category import LossCategory
import tensorflow as tf
from typing import Iterable, Tuple
from openfisca_core.taxscales import SingleAmountTaxScale


class IncomeTax(LossCategory):
    name = "Income Tax"
    category = "Programs"
    variable = "income_tax"

    def initialise(self):
        income_tax = self.sim.calc("income_tax", period=self.year).values
        total_income = self.sim.calc(
            "adjusted_net_income", period=self.year
        ).values
        countries = self.sim.calc("country", period=self.year).values
        household_income_tax = self.sim.calc(
            "income_tax", period=self.year, map_to="household"
        ).values

        it = self.calibration_parameters.programs.income_tax

        self.comparisons = []

        # Revenue by country

        for country in ("ENGLAND", "WALES", "NORTHERN_IRELAND", "SCOTLAND"):
            self.comparisons += [
                (
                    f"income_tax_{country}",
                    household_income_tax * (countries == country),
                    it.budgetary_impact.by_country._children[country],
                )
            ]

        self.comparisons += [
            (
                "income_tax_UNITED_KINGDOM",
                household_income_tax,
                it.budgetary_impact.by_country._children["UNITED_KINGDOM"],
            )
        ]

        # Revenue by income band

        scale = it.budgetary_impact.by_income

        for i in range(len(scale.thresholds)):
            lower_threshold = scale.thresholds[i]
            upper_threshold = (
                scale.thresholds[i + 1]
                if i < len(scale.thresholds) - 1
                else np.inf
            )

            income_is_in_band = (total_income >= lower_threshold) * (
                total_income < upper_threshold
            )
            household_values = self.sim.map_to(
                income_tax * income_is_in_band, "person", "household"
            )

            amount = scale.amounts[i]
            self.comparisons += [
                (
                    f"income_tax_by_income_{i}",
                    household_values,
                    amount,
                )
            ]

        # Taxpayers by country and income band

        tax_band = self.sim.calc("tax_band", period=self.year).values
        tax_band = np.select(
            [
                tax_band == "STARTER",
                tax_band == "INTERMEDIATE",
                True,
            ],
            [
                "BASIC",
                "BASIC",
                tax_band,
            ],
        )

        person_country = self.sim.calc(
            "country", period=self.year, map_to="person"
        ).values

        for country in ("ENGLAND", "WALES", "NORTHERN_IRELAND", "SCOTLAND"):
            for band in ("BASIC", "HIGHER", "ADDITIONAL"):
                self.comparisons += [
                    (
                        f"income_tax_payers_{country}_{band}",
                        self.sim.map_to(
                            (income_tax > 0)
                            * (person_country == country)
                            * (tax_band == band),
                            "person",
                            "household",
                        ),
                        it.participants.by_country_and_band._children[
                            country
                        ]._children[band],
                    )
                ]

    def get_loss_subcomponents(
        self, household_weights: tf.Tensor
    ) -> Iterable[Tuple]:
        for name, values, actual in self.comparisons:
            yield name, tf.reduce_sum(values * household_weights), actual

    def get_metric_names(self) -> Iterable[str]:
        return [x[0] for x in self.comparisons]



from openfisca_uk.model_api import *


@uprated(by="uprating.september_cpi")
class student_loans(Variable):
    value_type = float
    entity = Person
    label = "Student loans"
    definition_period = YEAR
    unit = GBP


@uprated(by="uprating.september_cpi")
class adult_ema(Variable):
    label = "Adult EMA"
    documentation = "Educational Maintenance Allowance for adults"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = GBP


@uprated(by="uprating.september_cpi")
class child_ema(Variable):
    label = "Child EMA"
    documentation = "Educational Maintenance Allowance for children"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = GBP


@uprated(by="uprating.september_cpi")
class access_fund(Variable):
    label = "Access Fund"
    documentation = "Access Fund for educational assistance"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = GBP


@uprated(by="uprating.september_cpi")
class education_grants(Variable):
    label = "Education grants"
    documentation = "Grants for education"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = GBP


class student_payments(Variable):
    value_type = float
    entity = Person
    label = "Student payments"
    definition_period = YEAR
    unit = GBP

    def formula(person, period, parameters):
        EDUCATION_BENEFITS = [
            "adult_ema",
            "child_ema",
            "access_fund",
            "education_grants",
        ]
        return add(person, period, EDUCATION_BENEFITS)



from openfisca_uk.model_api import *


class pension_credit_income(Variable):
    label = "Income for Pension Credit"
    entity = BenUnit
    definition_period = YEAR
    value_type = float
    unit = GBP
    reference = "https://www.legislation.gov.uk/ukpga/2002/16/section/15"

    def formula(benunit, period, parameters):
        sources = parameters(period).dwp.pension_credit.guarantee_credit.income
        total = add(benunit, period, sources)
        bi = parameters(period).contrib.ubi_center.basic_income
        if bi.interactions.include_in_means_tests:
            total += add(benunit, period, ["basic_income"])
        return max_(0, total)



from ...loss_category import LossCategory
import tensorflow as tf
from typing import Iterable, Tuple


class Households(LossCategory):
    name = "Households"
    category = "Demographics"

    def initialise(self):
        region_ct_band_parameter = (
            self.calibration_parameters.demographics.households.by_region_by_council_tax_band
        )

        self.comparisons = []

        ct_band = self.sim.calc("council_tax_band").values
        region = self.sim.calc("region").values

        # Region - CT band

        for target_region in region_ct_band_parameter._children:
            for target_ct_band in region_ct_band_parameter._children[
                target_region
            ]._children:
                self.comparisons.append(
                    (
                        f"{target_region}_{target_ct_band}",
                        (ct_band == target_ct_band)
                        & (region == target_region),
                        region_ct_band_parameter._children[
                            target_region
                        ]._children[target_ct_band],
                    )
                )

        # Region - tenure type

        tenure_type = self.sim.calc("ons_tenure_type").values

        region_tenure_parameter = (
            self.calibration_parameters.demographics.households.by_region_by_tenure_type
        )

        for target_region in region_tenure_parameter._children:
            for target_tenure_type in region_tenure_parameter._children[
                target_region
            ]._children:
                self.comparisons.append(
                    (
                        f"{target_region}_{target_tenure_type}",
                        (tenure_type == target_tenure_type)
                        & (region == target_region),
                        region_tenure_parameter._children[
                            target_region
                        ]._children[target_tenure_type],
                    )
                )

    def get_loss_subcomponents(
        self, household_weights: tf.Tensor
    ) -> Iterable[Tuple]:
        for name, values, actual in self.comparisons:
            yield (
                name,
                tf.reduce_sum(household_weights * values),
                actual,
            )

    def get_metric_names(self) -> Iterable[str]:
        return [x[0] for x in self.comparisons]
